{
  OSC for Propeller v1.0
  Copyright (c) 2015 Benjamin Yaroch

  OSC Specifications: opensoundcontrol.org/spec-1_0

  Reference Materials:
        http://cnmat.berkeley.edu/oscuino
        Github - https://github.com/CNMAT/OSC

  ///////////////////////////////////////////////////////////////////////////////////////////////////
  OSC Regular expression pattern matching rules from http://opensoundcontrol.org/spec-1_0
        1. '?' in the OSC Address Pattern matches any single character
        2. '*' in the OSC Address Pattern matches any sequence of zero or more characters
        3. A string of characters in square brackets (e.g., "[string]") in the OSC Address Pattern matches any character in the string.
                Inside square brackets, the minus sign (-) and exclamation point (!) have special meanings:
                two characters separated by a minus sign indicates the range of characters between the given two in ASCII collating sequence.
                An exclamation point at the beginning of a bracketed string negates the sense of the list, meaning that the list matches any
                character not in the list.
        4. A comma-separated list of strings enclosed in curly braces (e.g., "{foo,bar}") in the OSC Address Pattern matches any of the strings
                in the list.
        5. Any other character in an OSC Address Pattern can match only the same character.
  ///////////////////////////////////////////////////////////////////////////////////////////////////  
}
CON

  NULL                          = 0 

  addrBufferSize                = 1024
  typeTagBufferSize             = 64
  argBufferSize                 = 256

  ARG_TYPE_INT                  = "i"                   ' int32
  ARG_TYPE_FLOAT                = "f"                   ' float32
  ARG_TYPE_STRING               = "s"                   ' OSC-string
  ARG_TYPE_BLOB                 = "b"                   ' OSC-blob

VAR

  long  addrIdx
  long  argsIdx                                         ' Point to next type tag string buffer location 

  byte  addrBuffer[addrBufferSize]
  byte  typeTagBuffer[typeTagBufferSize]
  byte  argsBuffer[argBufferSize]

  byte  DecodeAddrBuffer[addrBufferSize]
  byte  DecodeTypeTagBuffer[typeTagBufferSize]
  byte  DecodeArgsBuffer[argBufferSize]
        
OBJ
 
  're   : "regex"

PUB init

  addrIdx := 0
  argsIdx := 0

  ' Clear Type Tag, Argument, & Address data buffer
  bytefill(@typeTagBuffer, NULL, typeTagBufferSize)                                                            
  bytefill(@argsBuffer, NULL, argBufferSize)                                    
  bytefill(@addrBuffer, NULL, addrBufferSize)                                   ' Zero Address buffer

PUB addressSet(_ptr) | stringPtr

  '' Sets/Appends the address this message will be sent to.

  if strsize(_ptr)+ addrIdx =< addrBufferSize                                   ' String must be shorter than buffer
    stringPtr := 0
    repeat strsize(_ptr)
      addrBuffer[addrIdx++] := byte[_ptr][stringPtr++]                          ' Transmit each byte in the string
     
    return @addrBuffer

PUB addressGet

  '' Return the address of received message.                

  return @DecodeAddrBuffer                              ' Pass the message buffer pointer                             

PUB typeGet(_typeIdx)

  '' Returns the type tag for a given location/index.

  return DecodeTypeTagBuffer[_typeIdx]
  
PUB argSet(_type, _data) | ptr, idx, data

  '' Sets/Appends a typed argument to this message.
  ''  NOTE: _data should be a string pointer for String and Blob types

  if argsIdx =< argBufferSize AND strsize(@typeTagBuffer) =< typeTagBufferSize
  
    case _type
      ARG_TYPE_INT:                                                             ' Integer
        typeTagBuffer[strsize(@typeTagBuffer)] := _type                         ' Append new Type tag

        argsBuffer[argsIdx++]   := _data.byte[3]
        argsBuffer[argsIdx++]   := _data.byte[2]
        argsBuffer[argsIdx++]   := _data.byte[1]
        argsBuffer[argsIdx++]   := _data.byte[0]
     
      ARG_TYPE_FLOAT:                                                           ' Float
        typeTagBuffer[strsize(@typeTagBuffer)] := _type                         ' Append new Type tag

        argsBuffer[argsIdx++]   := _data.byte[3]
        argsBuffer[argsIdx++]   := _data.byte[2]
        argsBuffer[argsIdx++]   := _data.byte[1]
        argsBuffer[argsIdx++]   := _data.byte[0]

      ARG_TYPE_STRING:                                                          ' String
        typeTagBuffer[strsize(@typeTagBuffer)] := _type                         ' Append new Type tag

        repeat idx from 0 to strsize(_data)                                   
          argsBuffer[argsIdx++] := byte[_data][idx]                             ' Copy string buffer, include NULL termination 
        
        repeat while argsIdx // 4                                               ' Is string 32bit aligned? Skip ahead to make string 32bit aligned
          argsIdx++
          
      ARG_TYPE_BLOB:                                                            ' Blob
        typeTagBuffer[strsize(@typeTagBuffer)] := _type                         ' Append new Type tag
        
        repeat idx from 0 to strsize(_data)                                  
          argsBuffer[argsIdx++] := byte[_data][idx]                             ' Copy string buffer, include NULL termination 
        
        repeat while argsIdx // 4                                               ' Is string 32bit aligned? Skip ahead to make string 32bit aligned
          argsIdx++
          
    return @argsBuffer

PUB argGet(_argIdx) | value, type, idx, ptr

  '' Get an argument from this message, based on location/index.
  ''  NOTE: result is a string pointer for String type and Blob type
  
  ptr := 0 

  repeat idx from 0 to _argIdx                                        
    case DecodeTypeTagBuffer[idx] 
      
      ARG_TYPE_INT:                                                           ' Integer
        value.byte[3] := DecodeArgsBuffer[ptr++]
        value.byte[2] := DecodeArgsBuffer[ptr++]
        value.byte[1] := DecodeArgsBuffer[ptr++]
        value.byte[0] := DecodeArgsBuffer[ptr++]

        result := value
     
      ARG_TYPE_FLOAT:                                                           ' FLOAT
        value.byte[3] := DecodeArgsBuffer[ptr++]
        value.byte[2] := DecodeArgsBuffer[ptr++]
        value.byte[1] := DecodeArgsBuffer[ptr++]    
        value.byte[0] := DecodeArgsBuffer[ptr++]

        result := value

      ARG_TYPE_STRING:                                                          ' STRING
        result := @DecodeArgsBuffer + ptr                                       ' Returns pointer to start of string
        
        repeat while DecodeArgsBuffer[ptr++] <> 0                               ' Advance string pointer                                                                                
        repeat while ptr // 4                                                   ' Is string 32bit aligned? Skip ahead to make string 32bit aligned
          ptr++
                                                                                
      ARG_TYPE_BLOB:                                                            ' BLOB
        result := @DecodeArgsBuffer + ptr                                       ' Returns pointer to start of string (aka blob)
        
        repeat while DecodeArgsBuffer[ptr++] <> 0                               ' Advance string pointer                                                                                
        repeat while ptr // 4                                                   ' Is string 32bit aligned? Skip ahead to make string 32bit aligned
          ptr++

PUB encode(_ptr) | idx, msgSize     

  '' Builds an OscMessage to provided buffer location.
  '' Returns: packet length, 0 if there was an error

  msgSize := 0

  if strsize(@addrBuffer) > 0                           ' OSC addresses cannot be empty

    ' --- Address string --- 
    repeat idx from 0 to strsize(@addrBuffer)                                 
      byte[_ptr][msgSize++] := addrBuffer[idx]          ' Copy string buffer, advance 1 to include NULL termination

    repeat while msgSize // 4                           ' Is string 32bit aligned? Skip ahead to make string 32bit aligned                           
      msgSize++

    ' --- Comma seperator ---         
    byte[_ptr][msgSize++] := ","                        ' Comma to indicate Type Tag string; advance string pointer
        
    ' --- Type Tag string ---
    repeat idx from 0 to strsize(@typeTagBuffer)
      byte[_ptr][msgSize++] := typeTagBuffer[idx]       ' Copy string buffer, advance 1 to include NULL termination

    repeat while msgSize // 4                           ' Is string 32bit aligned? Skip ahead to make string 32bit aligned                           
      msgSize++      

    ' --- Arguments string ---        
    repeat idx from 0 to argsIdx - 1
      byte[_ptr][msgSize++] := argsBuffer[idx]          ' Copy string buffer
    
    ' --- Clean Up ---         
    ' Reset pointers
    addrIdx := 0
    argsIdx := 0

    ' Clear all the data buffers
    bytefill(@typeTagBuffer, NULL, typeTagBufferSize)                                                          
    bytefill(@argsBuffer, NULL, argBufferSize)                                  
    bytefill(@addrBuffer, NULL, addrBufferSize)         ' Zero Address buffer

    return msgSize

  else
    return 0             

PUB decode(_ptr, _len) | msgPtr, idx

  '' Parses an OscMessage, stores to correct buffers.

  ' Clear all data buffers
  bytefill(@DecodeAddrBuffer, NULL, addrBufferSize)      ' Clear buffers before writing
  bytefill(@DecodeArgsBuffer, NULL, argBufferSize)        
  bytefill(@DecodeTypeTagBuffer, NULL, typetagBufferSize)
                                                          
  ' Reset pointers                                        
  msgPtr := 0                                             
                                                          
  ' --- Address string ---                                
  repeat idx from 0 to strsize(_ptr)                        
    DecodeAddrBuffer[msgPtr++] := byte[_ptr][msgPtr]     ' Copy string buffer, advance 1 to include NULL termination 

  repeat while msgPtr // 4                              ' Is string 32bit aligned? Skip ahead to make string 32bit aligned
    msgPtr++      

  ' --- Look for Type Tag comma seperator ---             
  if byte[_ptr][msgPtr++] == ","                         ' Comma to indicate Type Tag string; advance string pointer

    ' --- Move Type Tags ---                              
    idx := 0                                              
    repeat while byte[_ptr][msgPtr] <> 0                  
      DecodeTypeTagBuffer[idx++] := byte[_ptr][msgPtr++] ' Copy string buffer, advance 1 to include NULL termination

    repeat while msgPtr // 4                              ' Is string 32bit aligned? Skip ahead to make string 32bit aligned
      msgPtr++                      
                                                          
    ' --- Move Arguments ---                                   
    idx := 0                                            
    repeat until msgPtr == _len                          
      DecodeArgsBuffer[idx++] := byte[_ptr][msgPtr++]  
                                                          
    return true                                          ' Indicate that there were type tags
                                                          
  else                                                    
    return false                                         ' Indicate that there were no type tags

PUB fullMatch(_sourcePtr, _comparePtr) | idx

  '' Address full match
  '' Returns: True only for a complete/perfect match of address.
  
  result := true                                        ' initilize match status to true
                                  
  repeat idx from 0 to strsize(_sourcePtr) - 1
    if byte[_sourcePtr][idx] <> byte[_comparePtr][idx]  ' Compare recieved address with sample
      result := false          

PUB match(_sourcePtr, _comparePtr)

  '' Allows for partial matches. It returns the number of characters matched.
  '' Note that maches must match to the end or the / character.

  ' NOT IMPLEMENTED - FUTURE
  